#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access
import os
from abc import abstractmethod

from neptune.new.exceptions import (
    MetadataInconsistency,
    MissingFieldException,
    NeptuneOfflineModeFetchException,
    TypeDoesNotSupportAttributeException,
)


class AbstractExperimentTestMixin:
    @staticmethod
    @abstractmethod
    def call_init(**kwargs):
        pass

    def test_incorrect_mode(self):
        with self.assertRaises(ValueError):
            self.call_init(mode="srtgj")

    def test_debug_mode(self):
        exp = self.call_init(mode="debug")
        exp["some/variable"] = 13
        self.assertEqual(13, exp["some/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    def test_offline_mode(self):
        exp = self.call_init(mode="offline")
        exp["some/variable"] = 13
        with self.assertRaises(NeptuneOfflineModeFetchException):
            exp["some/variable"].fetch()

        exp_dir = f"{exp.container_type.value}__{exp._id}"
        self.assertIn(exp_dir, os.listdir(".neptune/offline"))
        self.assertIn("data-1.log", os.listdir(f".neptune/offline/{exp_dir}"))

    def test_sync_mode(self):
        exp = self.call_init(mode="sync")
        exp["some/variable"] = 13
        exp["copied/variable"] = exp["some/variable"]
        self.assertEqual(13, exp["some/variable"].fetch())
        self.assertEqual(13, exp["copied/variable"].fetch())
        self.assertNotIn(str(exp._id), os.listdir(".neptune"))

    def test_async_mode(self):
        with self.call_init(mode="async", flush_period=0.5) as exp:
            exp["some/variable"] = 13
            exp["copied/variable"] = exp["some/variable"]
            with self.assertRaises(MetadataInconsistency):
                exp["some/variable"].fetch()
            exp.wait()
            self.assertEqual(13, exp["some/variable"].fetch())
            self.assertEqual(13, exp["copied/variable"].fetch())

            exp_dir = f"{exp.container_type.value}__{exp._id}"
            self.assertIn(exp_dir, os.listdir(".neptune/async"))
            execution_dir = os.listdir(f".neptune/async/{exp_dir}")[0]
            self.assertIn(
                "data-1.log",
                os.listdir(f".neptune/async/{exp_dir}/{execution_dir}"),
            )

    def test_missing_attribute(self):
        exp = self.call_init(mode="debug")
        with self.assertRaises(MissingFieldException):
            exp["non/existing/path"].fetch()

    def test_wrong_function(self):
        exp = self.call_init(mode="debug")
        with self.assertRaises(AttributeError):
            exp["non/existing/path"].foo()

    def test_wrong_per_type_function(self):
        exp = self.call_init(mode="debug")
        exp["some/path"] = "foo"
        with self.assertRaises(TypeDoesNotSupportAttributeException):
            exp["some/path"].download()

    @abstractmethod
    def test_read_only_mode(self):
        pass



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from unittest.mock import MagicMock

import pytest

from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.operation import Operation
from neptune.new.sync import SyncRunner
from neptune.new.sync.utils import get_qualified_name
from tests.neptune.new.sync.utils import (
    execute_operations,
    generate_get_metadata_container,
    prepare_deprecated_run,
    prepare_metadata_container,
)


@pytest.fixture(name="backend")
def backend_fixture():
    backend = MagicMock()
    backend.execute_operations.side_effect = execute_operations
    return backend


@pytest.fixture(name="sync_runner")
def sync_runner_fixture(backend):
    return SyncRunner(backend=backend)


@pytest.mark.parametrize("container_type", list(ContainerType))
def test_sync_all_runs(tmp_path, mocker, capsys, backend, sync_runner, container_type):
    # given
    unsynced_container = prepare_metadata_container(
        container_type=container_type, path=tmp_path, last_ack_version=1
    )
    synced_container = prepare_metadata_container(
        container_type=container_type, path=tmp_path, last_ack_version=3
    )
    get_container_impl = generate_get_metadata_container(
        registered_containers=(unsynced_container, synced_container)
    )

    # and
    mocker.patch.object(backend, "get_metadata_container", get_container_impl)
    mocker.patch.object(Operation, "from_dict", lambda x: x)

    # when
    sync_runner.sync_all_containers(tmp_path, "foo")

    # then
    captured = capsys.readouterr()
    assert captured.err == ""

    # expect output for unsynced run
    assert f"Synchronising {get_qualified_name(unsynced_container)}" in captured.out
    assert (
        f"Synchronization of {container_type.value} {get_qualified_name(unsynced_container)} completed."
        in captured.out
    )

    # expect NO output for synced run
    assert f"Synchronising {get_qualified_name(synced_container)}" not in captured.out

    # and
    backend.execute_operations.has_calls(
        [
            mocker.call(unsynced_container.id, ContainerType.RUN, ["op-1", "op-2"]),
        ],
        any_order=True,
    )


def test_sync_all_offline_runs(tmp_path, mocker, capsys, backend, sync_runner):
    # given
    offline_run = prepare_metadata_container(
        container_type=ContainerType.RUN, path=tmp_path, last_ack_version=None
    )
    get_run_impl = generate_get_metadata_container(registered_containers=(offline_run,))

    # and
    mocker.patch.object(backend, "get_metadata_container", get_run_impl)
    mocker.patch.object(
        sync_runner,
        "_register_offline_run",
        lambda project, container_type: offline_run,
    )
    mocker.patch.object(Operation, "from_dict", lambda x: x)

    # when
    sync_runner.sync_all_containers(tmp_path, "foo")

    # then
    captured = capsys.readouterr()
    assert captured.err == ""
    assert (
        "Offline run run__{} registered as {}".format(
            f"{offline_run.id}", get_qualified_name(offline_run)
        )
    ) in captured.out

    # and
    backend.execute_operations.has_calls(
        [
            mocker.call(offline_run.id, ContainerType.RUN, ["op-1", "op-2"]),
        ],
        any_order=True,
    )


def test_sync_selected_runs(tmp_path, mocker, capsys, backend, sync_runner):
    # given
    unsync_exp = prepare_metadata_container(
        container_type=ContainerType.RUN, path=tmp_path, last_ack_version=1
    )  # won't be synced, despite fact it's not synced yet
    sync_exp = prepare_metadata_container(
        container_type=ContainerType.RUN, path=tmp_path, last_ack_version=3
    )  # will be synced despite fact that it's up to date
    offline_run = prepare_metadata_container(
        container_type=ContainerType.RUN, path=tmp_path, last_ack_version=None
    )  # will be synced
    get_run_impl = generate_get_metadata_container(
        registered_containers=[
            unsync_exp,
            sync_exp,
            offline_run,
        ]
    )

    # and
    mocker.patch.object(backend, "get_metadata_container", get_run_impl)
    mocker.patch.object(
        sync_runner,
        "_register_offline_run",
        lambda project, container_type: offline_run,
    )
    mocker.patch.object(Operation, "from_dict", lambda x: x)

    # when
    sync_runner.sync_selected_containers(
        base_path=tmp_path,
        project_name="some-name",
        container_names=[
            get_qualified_name(sync_exp),
            "offline/run__" + offline_run.id,
        ],
    )

    # then
    captured = capsys.readouterr()
    # expect no errors
    assert captured.err == ""

    # expected output for mentioned async exp
    assert "Synchronising {}".format(get_qualified_name(sync_exp)) in captured.out
    assert (
        "Synchronization of run {} completed.".format(get_qualified_name(sync_exp))
        in captured.out
    )

    # expected output for offline container
    assert (
        "Offline run run__{} registered as {}".format(
            f"{offline_run.id}", get_qualified_name(offline_run)
        )
    ) in captured.out
    assert "Synchronising {}".format(get_qualified_name(offline_run)) in captured.out
    assert (
        "Synchronization of run {} completed.".format(get_qualified_name(offline_run))
        in captured.out
    )

    # expected NO output for not mentioned async container
    assert "Synchronising {}".format(get_qualified_name(unsync_exp)) not in captured.out

    # and
    backend.execute_operations.has_calls(
        [
            mocker.call(
                sync_exp.id,
                ContainerType.RUN,
                operations=["op-1", "op-2"],
            ),
            mocker.call(
                offline_run.id,
                ContainerType.RUN,
                operations=["op-0", "op-1", "op-2"],
            ),
        ],
        any_order=True,
    )


def test_sync_deprecated_runs(tmp_path, mocker, capsys, backend, sync_runner):
    # given
    deprecated_unsynced_run = prepare_deprecated_run(path=tmp_path, last_ack_version=1)
    offline_old_run = prepare_deprecated_run(path=tmp_path, last_ack_version=None)
    get_container_impl = generate_get_metadata_container(
        registered_containers=(deprecated_unsynced_run, offline_old_run)
    )

    # and
    mocker.patch.object(backend, "get_metadata_container", get_container_impl)
    mocker.patch.object(
        sync_runner,
        "_register_offline_run",
        lambda project, container_type: offline_old_run,
    )
    mocker.patch.object(Operation, "from_dict", lambda x: x)

    # when
    sync_runner.sync_all_containers(tmp_path, "foo")

    # then
    captured = capsys.readouterr()
    assert captured.err == ""

    assert (
        "Offline run {} registered as {}".format(
            f"{offline_old_run.id}", get_qualified_name(offline_old_run)
        )
    ) in captured.out

    assert (
        "Synchronising {}".format(get_qualified_name(deprecated_unsynced_run))
        in captured.out
    )
    assert (
        "Synchronization of run {} completed.".format(
            get_qualified_name(deprecated_unsynced_run)
        )
        in captured.out
    )
    assert (
        "Synchronising {}".format(get_qualified_name(offline_old_run)) in captured.out
    )
    assert (
        "Synchronization of run {} completed.".format(
            get_qualified_name(offline_old_run)
        )
        in captured.out
    )

    # and
    backend.execute_operations.has_calls(
        [
            mocker.call(
                deprecated_unsynced_run.id,
                ContainerType.RUN,
                operations=["op-1", "op-2"],
            ),
            mocker.call(
                offline_old_run.id,
                ContainerType.RUN,
                operations=["op-0", "op-1", "op-2"],
            ),
        ],
        any_order=True,
    )


def test_sync_non_existent_container(tmp_path, capsys, sync_runner):
    # when
    sync_runner.sync_selected_containers(
        base_path=tmp_path, project_name="foo", container_names=["bar"]
    )

    # then
    captured = capsys.readouterr()
    assert "Warning: Run 'bar' does not exist in location" in captured.err


def test_sync_non_existent_offline_containers(tmp_path, capsys, sync_runner):
    # when
    sync_runner.sync_selected_containers(
        base_path=tmp_path, project_name="foo", container_names=["offline/foo__bar"]
    )
    sync_runner.sync_selected_containers(
        base_path=tmp_path, project_name="foo", container_names=["offline/model__bar"]
    )

    # then
    captured = capsys.readouterr()
    assert "Offline run foo__bar not found on disk." in captured.err
    assert "Offline run model__bar not found on disk." in captured.err



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import datetime
import unittest
import uuid
from random import randint
from time import time

from neptune.new.exceptions import ContainerUUIDNotFound, MetadataInconsistency
from neptune.new.internal.backends.api_model import (
    DatetimeAttribute,
    FloatAttribute,
    FloatPointValue,
    FloatSeriesAttribute,
    FloatSeriesValues,
    StringAttribute,
    StringPointValue,
    StringSeriesAttribute,
    StringSeriesValues,
    StringSetAttribute,
)
from neptune.new.internal.backends.neptune_backend_mock import NeptuneBackendMock
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.operation import (
    AddStrings,
    AssignDatetime,
    AssignFloat,
    AssignString,
    LogFloats,
    LogStrings,
)
from tests.neptune.random_utils import a_string


class TestNeptuneBackendMock(unittest.TestCase):
    # pylint:disable=protected-access

    def setUp(self) -> None:
        self.backend = NeptuneBackendMock()
        project_id = self.backend._project_id
        exp = self.backend.create_run(project_id=project_id)
        model = self.backend.create_model(
            project_id=project_id,
            key="MOD",
        )
        model_version = self.backend.create_model_version(
            project_id=project_id, model_id=model.id
        )
        self.ids_with_types = [
            (self.backend._project_id, ContainerType.PROJECT),
            (exp.id, ContainerType.RUN),
            (model.id, ContainerType.MODEL),
            (model_version.id, ContainerType.MODEL_VERSION),
        ]

    def test_get_float_attribute(self):
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                # given
                digit = randint(1, 10**4)
                self.backend.execute_operations(
                    container_id, container_type, operations=[AssignFloat(["x"], digit)]
                )

                # when
                ret = self.backend.get_float_attribute(
                    container_id, container_type, path=["x"]
                )

                # then
                self.assertEqual(FloatAttribute(digit), ret)

    def test_get_string_attribute(self):
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                # given
                text = a_string()
                self.backend.execute_operations(
                    container_id, container_type, operations=[AssignString(["x"], text)]
                )

                # when
                ret = self.backend.get_string_attribute(
                    container_id, container_type, path=["x"]
                )

                # then
                self.assertEqual(StringAttribute(text), ret)

    def test_get_datetime_attribute(self):
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                # given
                now = datetime.datetime.now()
                now = now.replace(microsecond=1000 * int(now.microsecond / 1000))
                self.backend.execute_operations(
                    container_id, container_type, [AssignDatetime(["x"], now)]
                )

                # when
                ret = self.backend.get_datetime_attribute(
                    container_id, container_type, ["x"]
                )

                # then
                self.assertEqual(DatetimeAttribute(now), ret)

    def test_get_float_series_attribute(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogFloats(
                            ["x"],
                            [
                                LogFloats.ValueType(5, None, time()),
                                LogFloats.ValueType(3, None, time()),
                            ],
                        )
                    ],
                )
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogFloats(
                            ["x"],
                            [
                                LogFloats.ValueType(2, None, time()),
                                LogFloats.ValueType(9, None, time()),
                            ],
                        )
                    ],
                )

                # when
                ret = self.backend.get_float_series_attribute(
                    container_id, container_type, ["x"]
                )

                # then
                self.assertEqual(FloatSeriesAttribute(9), ret)

    def test_get_string_series_attribute(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogStrings(
                            ["x"],
                            [
                                LogStrings.ValueType("adf", None, time()),
                                LogStrings.ValueType("sdg", None, time()),
                            ],
                        )
                    ],
                )
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogStrings(
                            ["x"],
                            [
                                LogStrings.ValueType("dfh", None, time()),
                                LogStrings.ValueType("qwe", None, time()),
                            ],
                        )
                    ],
                )

                # when
                ret = self.backend.get_string_series_attribute(
                    container_id, container_type, ["x"]
                )

                # then
                self.assertEqual(StringSeriesAttribute("qwe"), ret)

    def test_get_string_set_attribute(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AddStrings(["x"], {"abcx", "qwe"})]
                )

                # when
                ret = self.backend.get_string_set_attribute(
                    container_id, container_type, ["x"]
                )

                # then
        self.assertEqual(StringSetAttribute({"abcx", "qwe"}), ret)

    def test_get_string_series_values(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogStrings(
                            ["x"],
                            [
                                LogStrings.ValueType("adf", None, time()),
                                LogStrings.ValueType("sdg", None, time()),
                            ],
                        )
                    ],
                )
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogStrings(
                            ["x"],
                            [
                                LogStrings.ValueType("dfh", None, time()),
                                LogStrings.ValueType("qwe", None, time()),
                            ],
                        )
                    ],
                )

                # when
                ret = self.backend.get_string_series_values(
                    container_id, container_type, path=["x"], limit=100, offset=0
                )

                # then
                self.assertEqual(
                    StringSeriesValues(
                        4,
                        [
                            StringPointValue(
                                timestampMillis=42342, step=0, value="adf"
                            ),
                            StringPointValue(
                                timestampMillis=42342, step=1, value="sdg"
                            ),
                            StringPointValue(
                                timestampMillis=42342, step=2, value="dfh"
                            ),
                            StringPointValue(
                                timestampMillis=42342, step=3, value="qwe"
                            ),
                        ],
                    ),
                    ret,
                )

    def test_get_float_series_values(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogFloats(
                            ["x"],
                            [
                                LogFloats.ValueType(5, None, time()),
                                LogFloats.ValueType(3, None, time()),
                            ],
                        )
                    ],
                )
                self.backend.execute_operations(
                    container_id,
                    container_type,
                    [
                        LogFloats(
                            ["x"],
                            [
                                LogFloats.ValueType(2, None, time()),
                                LogFloats.ValueType(9, None, time()),
                            ],
                        )
                    ],
                )

                # when
                ret = self.backend.get_float_series_values(
                    container_id, container_type, path=["x"], limit=100, offset=0
                )

                # then
                self.assertEqual(
                    FloatSeriesValues(
                        4,
                        [
                            FloatPointValue(timestampMillis=42342, step=0, value=5),
                            FloatPointValue(timestampMillis=42342, step=1, value=3),
                            FloatPointValue(timestampMillis=42342, step=2, value=2),
                            FloatPointValue(timestampMillis=42342, step=3, value=9),
                        ],
                    ),
                    ret,
                )

    def test_get_float_attribute_wrong_type(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AssignString(["x"], "abc")]
                )

                # then
                with self.assertRaises(MetadataInconsistency):
                    self.backend.get_float_series_attribute(
                        container_id, container_type, ["x"]
                    )

    def test_get_string_attribute_wrong_type(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AssignFloat(["x"], 5)]
                )

                # then
                with self.assertRaises(MetadataInconsistency):
                    self.backend.get_string_attribute(
                        container_id, container_type, ["x"]
                    )

    def test_get_datetime_attribute_wrong_type(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AssignString(["x"], "abc")]
                )

                # then
                with self.assertRaises(MetadataInconsistency):
                    self.backend.get_datetime_attribute(
                        container_id, container_type, ["x"]
                    )

    def test_get_string_series_attribute_wrong_type(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AssignString(["x"], "abc")]
                )

                # then
                with self.assertRaises(MetadataInconsistency):
                    self.backend.get_string_series_attribute(
                        container_id, container_type, ["x"]
                    )

    def test_get_string_set_attribute_wrong_type(self):
        # given
        for container_id, container_type in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AssignString(["x"], "abc")]
                )

                # then
                with self.assertRaises(MetadataInconsistency):
                    self.backend.get_string_set_attribute(
                        container_id, container_type, ["x"]
                    )

    def test_container_not_found(self):
        # given
        for (container_id, container_type) in self.ids_with_types:
            with self.subTest(f"For containerType: {container_type}"):
                self.backend.execute_operations(
                    container_id, container_type, [AssignString(["x"], "abc")]
                )

                # then
                with self.assertRaises(ContainerUUIDNotFound):
                    self.backend.get_float_series_attribute(
                        str(uuid.uuid4()), container_type, ["x"]
                    )



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import concurrent.futures
import random

import pytest

import neptune.new as neptune
from e2e_tests.base import BaseE2ETest, fake
from e2e_tests.utils import reinitialize_container


def store_in_container(
    sys_id: str, project: str, container_type: str, destination: str
):
    container = reinitialize_container(
        sys_id=sys_id, container_type=container_type, project=project
    )
    container[destination] = fake.color()
    container.sync()


class TestMultiple(BaseE2ETest):
    @pytest.mark.parametrize(
        "container", ["run", "model", "model_version"], indirect=True
    )
    def test_single_thread(
        self, container: neptune.metadata_containers.MetadataContainer, environment
    ):
        sys_id = container["sys/id"].fetch()
        number_of_reinitialized = 5
        namespace = self.gen_key()

        reinitialized = [
            reinitialize_container(
                sys_id=sys_id,
                container_type=container.container_type.value,
                project=environment.project,
            )
            for _ in range(number_of_reinitialized)
        ]

        container[f"{namespace}/{fake.unique.word()}"] = fake.color()
        container.sync()

        random.shuffle(reinitialized)
        for reinitialized_container in reinitialized:
            reinitialized_container[f"{namespace}/{fake.unique.word()}"] = fake.color()

        random.shuffle(reinitialized)
        for reinitialized_container in reinitialized:
            reinitialized_container.sync()

        container.sync()

        assert len(container[namespace].fetch()) == number_of_reinitialized + 1

        for r in reinitialized:
            r.stop()

    @pytest.mark.skip(reason="no way of currently testing this")
    @pytest.mark.parametrize(
        "container", ["run", "model", "model_version"], indirect=True
    )
    def test_multiple_processes(self, container: neptune.Run, environment):
        number_of_reinitialized = 10
        namespace = self.gen_key()

        container[f"{namespace}/{fake.unique.word()}"] = fake.color()

        with concurrent.futures.ProcessPoolExecutor(max_workers=3) as executor:
            futures = [
                executor.submit(
                    store_in_container,
                    sys_id=container["sys/id"].fetch(),
                    container_type=container.container_type.value,
                    project=environment.project,
                    destination=f"{namespace}/{fake.unique.word()}",
                )
                for _ in range(number_of_reinitialized)
            ]
            for future in concurrent.futures.as_completed(futures):
                _ = future.result()

        container.sync()

        assert len(container[namespace].fetch()) == number_of_reinitialized + 1

    @pytest.mark.parametrize(
        "container", ["run", "model", "model_version"], indirect=True
    )
    def test_multiple_threads(self, container: neptune.Run, environment):
        number_of_reinitialized = 10
        namespace = self.gen_key()

        container[f"{namespace}/{fake.unique.word()}"] = fake.color()

        with concurrent.futures.ThreadPoolExecutor(max_workers=3) as executor:
            futures = [
                executor.submit(
                    store_in_container,
                    sys_id=container["sys/id"].fetch(),
                    container_type=container.container_type.value,
                    project=environment.project,
                    destination=f"{namespace}/{fake.unique.word()}",
                )
                for _ in range(number_of_reinitialized)
            ]
            for future in concurrent.futures.as_completed(futures):
                _ = future.result()

        container.sync()

        assert len(container[namespace].fetch()) == number_of_reinitialized + 1



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import itertools
import random

import pytest

from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from neptune.new.metadata_containers import MetadataContainer

# List of every possible container type pair for instance: "run-run, run-model, model-model_version, ..."
ALL_CONTAINERS_PAIRS = list(
    map("-".join, itertools.product(AVAILABLE_CONTAINERS, AVAILABLE_CONTAINERS))
)


class TestCopying(BaseE2ETest):
    @pytest.mark.parametrize("containers_pair", ALL_CONTAINERS_PAIRS, indirect=True)
    @pytest.mark.parametrize(
        "value", [random.randint(0, 100), random.random(), fake.boolean(), fake.word()]
    )
    def test_copy(self, containers_pair: (MetadataContainer, MetadataContainer), value):
        container_a, container_b = containers_pair

        src, destination, destination2 = self.gen_key(), self.gen_key(), self.gen_key()

        container_a[src] = value
        container_a.sync()

        container_b[destination] = container_a[src]
        container_b[destination2] = container_b[destination]
        container_b.sync()

        assert container_a[src].fetch() == value
        assert container_b[destination].fetch() == value
        assert container_b[destination2].fetch() == value



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import random
import time
import uuid
from datetime import datetime, timezone

import pytest

import neptune.new as neptune
from e2e_tests.base import AVAILABLE_CONTAINERS, BaseE2ETest, fake
from neptune.new.metadata_containers import MetadataContainer, Model


class TestAtoms(BaseE2ETest):
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    @pytest.mark.parametrize(
        "value", [random.randint(0, 100), random.random(), fake.boolean(), fake.word()]
    )
    def test_simple_assign_and_fetch(self, container: MetadataContainer, value):
        key = self.gen_key()

        container[key] = value
        container.sync()
        assert container[key].fetch() == value

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_simple_assign_datetime(self, container: MetadataContainer):
        key = self.gen_key()
        now = datetime.now()

        container[key] = now
        container.sync()

        # expect truncate to milliseconds and add UTC timezone
        expected_now = now.astimezone(timezone.utc).replace(
            microsecond=int(now.microsecond / 1000) * 1000
        )
        assert container[key].fetch() == expected_now

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_fetch_non_existing_key(self, container: MetadataContainer):
        key = self.gen_key()
        with pytest.raises(AttributeError):
            container[key].fetch()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_delete_atom(self, container: MetadataContainer):
        key = self.gen_key()
        value = fake.name()

        container[key] = value
        container.sync()

        assert container[key].fetch() == value

        del container[key]
        with pytest.raises(AttributeError):
            container[key].fetch()


class TestNamespace(BaseE2ETest):
    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_reassigning(self, container: MetadataContainer):
        namespace = self.gen_key()
        key = f"{fake.unique.word()}/{fake.unique.word()}"
        value = fake.name()

        # Assign a namespace
        container[namespace] = {f"{key}": value}
        container.sync()

        assert container[f"{namespace}/{key}"].fetch() == value

        # Direct reassign internal value
        value = fake.name()
        container[f"{namespace}/{key}"] = value
        container.sync()

        assert container[f"{namespace}/{key}"].fetch() == value

        # Reassigning by namespace
        value = fake.name()
        container[namespace] = {f"{key}": value}
        container.sync()

        assert container[f"{namespace}/{key}"].fetch() == value

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_distinct_types(self, container: MetadataContainer):
        namespace = self.gen_key()
        key = f"{fake.unique.word()}/{fake.unique.word()}"
        value = random.randint(0, 100)

        container[namespace] = {f"{key}": value}
        container.sync()

        assert container[f"{namespace}/{key}"].fetch() == value

        new_value = fake.name()

        with pytest.raises(ValueError):
            container[namespace] = {f"{key}": new_value}
            container.sync()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_delete_namespace(self, container: MetadataContainer):
        namespace = fake.unique.word()
        key1 = fake.unique.word()
        key2 = fake.unique.word()
        value1 = fake.name()
        value2 = fake.name()

        container[namespace][key1] = value1
        container[namespace][key2] = value2
        container.sync()

        assert container[namespace][key1].fetch() == value1
        assert container[namespace][key2].fetch() == value2

        del container[namespace]
        with pytest.raises(AttributeError):
            container[namespace][key1].fetch()
        with pytest.raises(AttributeError):
            container[namespace][key2].fetch()


class TestStringSet(BaseE2ETest):
    neptune_tags_path = "sys/tags"

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_do_not_accept_non_tag_path(self, container: MetadataContainer):
        random_path = "some/path"
        container[random_path].add(fake.unique.word())
        container.sync()

        with pytest.raises(AttributeError):
            # backends accepts `'sys/tags'` only
            container[random_path].fetch()

    @pytest.mark.parametrize("container", AVAILABLE_CONTAINERS, indirect=True)
    def test_add_and_remove_tags(self, container: MetadataContainer):
        remaining_tag1 = fake.unique.word()
        remaining_tag2 = fake.unique.word()
        to_remove_tag1 = fake.unique.word()
        to_remove_tag2 = fake.unique.word()

        container.sync()
        if container.exists(self.neptune_tags_path):
            container[self.neptune_tags_path].clear()
        container[self.neptune_tags_path].add(remaining_tag1)
        container[self.neptune_tags_path].add([to_remove_tag1, remaining_tag2])
        container[self.neptune_tags_path].remove(to_remove_tag1)
        container[self.neptune_tags_path].remove(
            to_remove_tag2
        )  # remove non existing tag
        container.sync()

        assert container[self.neptune_tags_path].fetch() == {
            remaining_tag1,
            remaining_tag2,
        }


class TestFetchTable(BaseE2ETest):
    def test_fetch_runs_table(self, environment):
        tag = str(uuid.uuid4())

        with neptune.init_run(project=environment.project) as run:
            run["sys/tags"].add(tag)
            run["value"] = 12
            run.sync()

        with neptune.init_run(project=environment.project) as run:
            run["sys/tags"].add(tag)
            run["another/value"] = "testing"
            run.sync()

        # wait for the elasticsearch cache to fill
        time.sleep(5)

        project = neptune.get_project(name=environment.project)

        runs_table = sorted(
            project.fetch_runs_table(tag=tag).to_rows(),
            key=lambda r: r.get_attribute_value("sys/id"),
        )
        assert len(runs_table) == 2
        assert runs_table[0].get_attribute_value("value") == 12
        assert runs_table[1].get_attribute_value("another/value") == "testing"

    @pytest.mark.parametrize("container", ["model"], indirect=True)
    def test_fetch_model_versions_table(self, container: Model, environment):
        model_sys_id = container["sys/id"].fetch()
        versions_to_initialize = 5

        for _ in range(versions_to_initialize):
            with neptune.init_model_version(
                model=model_sys_id, project=environment.project
            ):
                pass

        # wait for the elasticsearch cache to fill
        time.sleep(5)

        versions_table = sorted(
            container.fetch_model_versions_table().to_rows(),
            key=lambda r: r.get_attribute_value("sys/id"),
        )
        assert len(versions_table) == versions_to_initialize
        for index in range(versions_to_initialize):
            assert (
                versions_table[index].get_attribute_value("sys/id")
                == f"{model_sys_id}-{index + 1}"
            )



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["SyncRunner"]

import logging
import os
import sys
import threading
import time
from pathlib import Path
from typing import Iterable, List, Optional, Sequence

import click

from neptune.new.constants import (
    ASYNC_DIRECTORY,
    OFFLINE_DIRECTORY,
    OFFLINE_NAME_PREFIX,
)
from neptune.new.envs import NEPTUNE_SYNC_BATCH_TIMEOUT_ENV
from neptune.new.exceptions import (
    CannotSynchronizeOfflineRunsWithoutProject,
    NeptuneConnectionLostException,
)
from neptune.new.internal.backends.api_model import ApiExperiment, Project
from neptune.new.internal.container_type import ContainerType
from neptune.new.internal.disk_queue import DiskQueue
from neptune.new.internal.id_formats import QualifiedName, UniqueId
from neptune.new.internal.operation import Operation
from neptune.new.sync.abstract_backend_runner import AbstractBackendRunner
from neptune.new.sync.utils import (
    create_dir_name,
    get_metadata_container,
    get_offline_dirs,
    get_project,
    get_qualified_name,
    is_container_synced,
    iterate_containers,
    split_dir_name,
)

retries_timeout = int(os.getenv(NEPTUNE_SYNC_BATCH_TIMEOUT_ENV, "3600"))


class SyncRunner(AbstractBackendRunner):
    def sync_run(self, run_path: Path, run: ApiExperiment) -> None:
        qualified_run_name = get_qualified_name(run)
        click.echo("Synchronising {}".format(qualified_run_name))
        for execution_path in run_path.iterdir():
            self.sync_execution(
                execution_path=execution_path,
                container_id=run.id,
                container_type=run.type,
            )
        click.echo(
            f"Synchronization of {run.type.value} {qualified_run_name} completed."
        )

    def sync_execution(
        self,
        execution_path: Path,
        container_id: UniqueId,
        container_type: ContainerType,
    ) -> None:
        disk_queue = DiskQueue(
            dir_path=execution_path,
            to_dict=lambda x: x.to_dict(),
            from_dict=Operation.from_dict,
            lock=threading.RLock(),
        )
        while True:
            batch, version = disk_queue.get_batch(1000)
            if not batch:
                break

            start_time = time.monotonic()
            expected_count = len(batch)
            version_to_ack = version - expected_count
            while True:
                try:
                    processed_count, _ = self._backend.execute_operations(
                        container_id=container_id,
                        container_type=container_type,
                        operations=batch,
                    )
                    version_to_ack += processed_count
                    batch = batch[processed_count:]
                    disk_queue.ack(version)
                    if version_to_ack == version:
                        break
                except NeptuneConnectionLostException as ex:
                    if time.monotonic() - start_time > retries_timeout:
                        raise ex
                    click.echo(
                        "Experiencing connection interruptions. "
                        "Will try to reestablish communication with Neptune. "
                        f"Internal exception was: {ex.cause.__class__.__name__}",
                        sys.stderr,
                    )

    def sync_all_registered_containers(self, base_path: Path) -> None:
        async_path = base_path / ASYNC_DIRECTORY
        for container_type, unique_id, path in iterate_containers(async_path):
            if not is_container_synced(path):
                run = get_metadata_container(
                    backend=self._backend,
                    container_id=unique_id,
                    container_type=container_type,
                )
                if run:
                    self.sync_run(run_path=path, run=run)

    def sync_selected_registered_containers(
        self, base_path: Path, qualified_container_names: Sequence[QualifiedName]
    ) -> None:
        for name in qualified_container_names:
            run = get_metadata_container(
                backend=self._backend,
                container_id=name,
            )
            if run:
                run_path = (
                    base_path / ASYNC_DIRECTORY / f"{create_dir_name(run.type, run.id)}"
                )
                run_path_deprecated = base_path / ASYNC_DIRECTORY / f"{run.id}"
                if run_path.exists():
                    self.sync_run(run_path=run_path, run=run)
                elif run_path_deprecated.exists():
                    self.sync_run(run_path=run_path_deprecated, run=run)
                else:
                    click.echo(
                        "Warning: Run '{}' does not exist in location {}".format(
                            name, base_path
                        ),
                        file=sys.stderr,
                    )

    def _register_offline_run(
        self, project: Project, container_type: ContainerType
    ) -> Optional[ApiExperiment]:
        try:
            if container_type == ContainerType.RUN:
                return self._backend.create_run(project.id)
            else:
                raise ValueError("Only runs are supported in offline mode")
        except Exception as e:
            click.echo(
                "Exception occurred while trying to create a run "
                "on the Neptune server. Please try again later",
                file=sys.stderr,
            )
            logging.exception(e)
            return None

    @staticmethod
    def _move_offline_run(
        base_path: Path,
        offline_dir: str,
        server_id: UniqueId,
        server_type: ContainerType,
    ) -> None:
        online_dir = create_dir_name(container_type=server_type, container_id=server_id)
        # create async directory for run
        (base_path / ASYNC_DIRECTORY / online_dir).mkdir(parents=True)
        # mv offline directory inside async one
        (base_path / OFFLINE_DIRECTORY / offline_dir).rename(
            base_path / ASYNC_DIRECTORY / online_dir / "exec-0-offline"
        )

    def register_offline_runs(
        self, base_path: Path, project: Project, offline_dirs: Iterable[str]
    ) -> List[ApiExperiment]:
        result = []
        for offline_dir in offline_dirs:
            offline_path = base_path / OFFLINE_DIRECTORY / offline_dir
            if offline_path.is_dir():
                container_type, _ = split_dir_name(dir_name=offline_dir)
                run = self._register_offline_run(project, container_type=container_type)
                if run:
                    self._move_offline_run(
                        base_path=base_path,
                        offline_dir=offline_dir,
                        server_id=run.id,
                        server_type=run.type,
                    )
                    click.echo(
                        f"Offline run {offline_dir} registered as {get_qualified_name(run)}"
                    )
                    result.append(run)
            else:
                click.echo(
                    f"Offline run {offline_dir} not found on disk.",
                    err=True,
                )
        return result

    def sync_offline_runs(
        self,
        base_path: Path,
        project_name: Optional[QualifiedName],
        offline_dirs: Sequence[UniqueId],
    ):
        if offline_dirs:
            project = get_project(project_name, backend=self._backend)
            if not project:
                raise CannotSynchronizeOfflineRunsWithoutProject
            registered_runs = self.register_offline_runs(
                base_path, project, offline_dirs
            )
            offline_runs_names = [get_qualified_name(exp) for exp in registered_runs]
            self.sync_selected_registered_containers(base_path, offline_runs_names)

    def sync_selected_containers(
        self,
        base_path: Path,
        project_name: Optional[str],
        container_names: Sequence[str],
    ) -> None:
        non_offline_container_names = [
            QualifiedName(name)
            for name in container_names
            if not name.startswith(OFFLINE_NAME_PREFIX)
        ]
        self.sync_selected_registered_containers(base_path, non_offline_container_names)

        offline_dirs = [
            UniqueId(name[len(OFFLINE_NAME_PREFIX) :])
            for name in container_names
            if name.startswith(OFFLINE_NAME_PREFIX)
        ]
        self.sync_offline_runs(base_path, project_name, offline_dirs)

    def sync_all_containers(self, base_path: Path, project_name: Optional[str]) -> None:
        self.sync_all_registered_containers(base_path)

        offline_dirs = get_offline_dirs(base_path)
        self.sync_offline_runs(base_path, project_name, offline_dirs)



#
# Copyright (c) 2021, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
__all__ = [
    "with_check_if_file_appears",
    "tmp_context",
    "a_project_name",
    "a_key",
    "Environment",
    "initialize_container",
    "reinitialize_container",
]

import io
import os
import random
import string
import tempfile
from collections import namedtuple
from contextlib import contextmanager
from datetime import datetime

import numpy
from PIL import Image
from PIL.PngImagePlugin import PngImageFile

import neptune.new as neptune


def _remove_file_if_exists(filepath):
    try:
        os.remove(filepath)
    except OSError:
        pass


# init kwargs which significantly reduce operations noise
DISABLE_SYSLOG_KWARGS = {
    "capture_stdout": False,
    "capture_stderr": False,
    "capture_hardware_metrics": False,
}


@contextmanager
def with_check_if_file_appears(filepath):
    """Checks if file will be present when leaving the block.
    File is removed if exists when entering the block."""
    _remove_file_if_exists(filepath)

    try:
        yield
    finally:
        assert os.path.exists(filepath)
        _remove_file_if_exists(filepath)


@contextmanager
def preserve_cwd(path):
    cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(cwd)


@contextmanager
def tmp_context():
    with tempfile.TemporaryDirectory() as tmp:
        with preserve_cwd(tmp):
            yield tmp


def generate_image(*, size: int) -> Image:
    random_numbers = numpy.random.rand(size, size, 3) * 255
    return Image.fromarray(random_numbers.astype("uint8")).convert("RGB")


def image_to_png(*, image: Image) -> PngImageFile:
    png_buf = io.BytesIO()
    image.save(png_buf, format="png")
    png_buf.seek(0)
    return PngImageFile(png_buf)


def a_key():
    return "".join(random.choices(string.ascii_uppercase, k=10))


def a_project_name(project_slug: str):
    project_name = f"e2e-{datetime.now().strftime('%Y%m%d-%H%M')}-{project_slug}"
    project_key = a_key()

    return project_name, project_key


Environment = namedtuple(
    "Environment",
    ["workspace", "project", "user_token", "admin_token", "admin", "user"],
)


def initialize_container(container_type, project, **extra_args):
    if container_type == "project":
        return neptune.init_project(name=project, **extra_args)

    if container_type == "run":
        return neptune.init_run(project=project, **extra_args)

    if container_type == "model":
        return neptune.init_model(key=a_key(), project=project, **extra_args)

    if container_type == "model_version":
        model = neptune.init_model(key=a_key(), project=project, **extra_args)
        model_sys_id = model["sys/id"].fetch()
        model.stop()

        return neptune.init_model_version(
            model=model_sys_id, project=project, **extra_args
        )

    raise NotImplementedError(container_type)


def reinitialize_container(sys_id: str, container_type: str, project: str):
    if container_type == "project":
        # exactly same as initialize_container(project), for convenience
        return neptune.init_project(name=project)

    if container_type == "run":
        return neptune.init_run(run=sys_id, project=project)

    if container_type == "model":
        return neptune.init_model(model=sys_id, project=project)

    if container_type == "model_version":
        return neptune.init_model_version(version=sys_id, project=project)

    raise NotImplementedError()



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import unicode_literals

import threading
import time
import unittest

import mock
import pytest

from neptune.internal.channels.channels import (
    ChannelIdWithValues,
    ChannelNamespace,
    ChannelType,
    ChannelValue,
)
from neptune.internal.channels.channels_values_sender import (
    ChannelsValuesSender,
    ChannelsValuesSendingThread,
)
from tests.neptune.experiments_object_factory import a_channel


class TestChannelsValuesSender(unittest.TestCase):
    _TS = time.time()

    _EXPERIMENT = mock.MagicMock()

    _NUMERIC_CHANNEL = a_channel()
    _NUMERIC_CHANNEL.update(name="1numeric", channelType=ChannelType.NUMERIC.value)

    _TEXT_CHANNEL = a_channel()
    _TEXT_CHANNEL.update(name="2text", channelType=ChannelType.TEXT.value)

    _IMAGE_CHANNEL = a_channel()
    _IMAGE_CHANNEL.update(name="3image", channelType=ChannelType.IMAGE.value)

    _OTHER_CHANNEL = a_channel()

    _CHANNELS = {
        c.name: c
        for c in [_NUMERIC_CHANNEL, _TEXT_CHANNEL, _IMAGE_CHANNEL, _OTHER_CHANNEL]
    }

    # pylint: disable=protected-access
    _BATCH_SIZE = ChannelsValuesSendingThread._MAX_VALUES_BATCH_LENGTH
    # pylint: disable=protected-access
    _IMAGES_BATCH_IMAGE_SIZE = (
        ChannelsValuesSendingThread._MAX_IMAGE_VALUES_BATCH_SIZE / 3
    )
    _IMAGES_BATCH_SIZE = 3

    def setUp(self):
        # pylint: disable=protected-access
        self._EXPERIMENT._get_channels.return_value = self._CHANNELS

        # pylint: disable=unused-argument
        def create_channel_fun(
            channel_name, channel_type, channel_namespace=ChannelNamespace.USER
        ):
            if channel_name == self._NUMERIC_CHANNEL.name:
                return self._NUMERIC_CHANNEL
            if channel_name == self._TEXT_CHANNEL.name:
                return self._TEXT_CHANNEL
            if channel_name == self._IMAGE_CHANNEL.name:
                return self._IMAGE_CHANNEL
            if channel_name == self._OTHER_CHANNEL.name:
                return self._OTHER_CHANNEL
            raise ValueError("unexpected channel")

        self._EXPERIMENT._create_channel = mock.MagicMock(
            side_effect=create_channel_fun
        )

    def tearDown(self):
        self._EXPERIMENT.reset_mock()

    def test_send_values_on_join(self):
        # given
        channel_value = ChannelValue(x=1, y="value", ts=self._TS)
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        channels_values_sender.send(
            self._TEXT_CHANNEL.name, self._TEXT_CHANNEL.channelType, channel_value
        )
        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        self._EXPERIMENT._send_channels_values.assert_called_with(
            [
                ChannelIdWithValues(
                    channel_id=self._TEXT_CHANNEL.id,
                    channel_name=self._TEXT_CHANNEL.name,
                    channel_type=self._TEXT_CHANNEL.channelType,
                    channel_namespace=ChannelNamespace.USER,
                    channel_values=[channel_value],
                )
            ]
        )

    def test_send_values_in_multiple_batches(self):
        # given
        channels_values = [
            ChannelValue(x=i, y="value{}".format(i), ts=self._TS + i)
            for i in range(0, self._BATCH_SIZE * 3)
        ]
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in channels_values:
            channels_values_sender.send(
                self._TEXT_CHANNEL.name, self._TEXT_CHANNEL.channelType, channel_value
            )
        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        self.assertEqual(
            self._EXPERIMENT._send_channels_values.mock_calls,
            [
                mock.call._send_channels_values(
                    [
                        ChannelIdWithValues(
                            channel_id=self._TEXT_CHANNEL.id,
                            channel_name=self._TEXT_CHANNEL.name,
                            channel_type=self._TEXT_CHANNEL.channelType,
                            channel_namespace=ChannelNamespace.USER,
                            channel_values=channels_values[0 : self._BATCH_SIZE],
                        )
                    ]
                ),
                mock.call._send_channels_values(
                    [
                        ChannelIdWithValues(
                            channel_id=self._TEXT_CHANNEL.id,
                            channel_name=self._TEXT_CHANNEL.name,
                            channel_type=self._TEXT_CHANNEL.channelType,
                            channel_namespace=ChannelNamespace.USER,
                            channel_values=channels_values[
                                self._BATCH_SIZE : self._BATCH_SIZE * 2
                            ],
                        )
                    ]
                ),
                mock.call._send_channels_values(
                    [
                        ChannelIdWithValues(
                            channel_id=self._TEXT_CHANNEL.id,
                            channel_name=self._TEXT_CHANNEL.name,
                            channel_type=self._TEXT_CHANNEL.channelType,
                            channel_namespace=ChannelNamespace.USER,
                            channel_values=channels_values[
                                self._BATCH_SIZE * 2 : self._BATCH_SIZE * 3
                            ],
                        )
                    ]
                ),
            ],
        )

    def test_send_images_in_smaller_batches(self):
        # and
        value = "base64Image=="
        channels_values = [
            ChannelValue(
                x=i,
                y={
                    "image_value": {
                        "data": value
                        + value * int(self._IMAGES_BATCH_IMAGE_SIZE / (len(value)))
                    }
                },
                ts=self._TS + i,
            )
            for i in range(0, self._IMAGES_BATCH_SIZE * 3)
        ]
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in channels_values:
            channels_values_sender.send(
                self._IMAGE_CHANNEL.name, self._IMAGE_CHANNEL.channelType, channel_value
            )
        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        self.assertEqual(
            self._EXPERIMENT._send_channels_values.mock_calls,
            [
                mock.call._send_channels_values(
                    [
                        ChannelIdWithValues(
                            channel_id=self._IMAGE_CHANNEL.id,
                            channel_name=self._IMAGE_CHANNEL.name,
                            channel_type=self._IMAGE_CHANNEL.channelType,
                            channel_namespace=ChannelNamespace.USER,
                            channel_values=channels_values[0 : self._IMAGES_BATCH_SIZE],
                        )
                    ]
                ),
                mock.call._send_channels_values(
                    [
                        ChannelIdWithValues(
                            channel_id=self._IMAGE_CHANNEL.id,
                            channel_name=self._IMAGE_CHANNEL.name,
                            channel_type=self._IMAGE_CHANNEL.channelType,
                            channel_namespace=ChannelNamespace.USER,
                            channel_values=channels_values[
                                self._IMAGES_BATCH_SIZE : self._IMAGES_BATCH_SIZE * 2
                            ],
                        )
                    ]
                ),
                mock.call._send_channels_values(
                    [
                        ChannelIdWithValues(
                            channel_id=self._IMAGE_CHANNEL.id,
                            channel_name=self._IMAGE_CHANNEL.name,
                            channel_type=self._IMAGE_CHANNEL.channelType,
                            channel_namespace=ChannelNamespace.USER,
                            channel_values=channels_values[
                                self._IMAGES_BATCH_SIZE * 2 :
                            ],
                        )
                    ]
                ),
            ],
        )

    def test_send_values_from_multiple_channels(self):
        # given
        numeric_values = [ChannelValue(x=i, y=i, ts=self._TS + i) for i in range(0, 3)]

        text_values = [
            ChannelValue(x=i, y="text", ts=self._TS + i) for i in range(0, 3)
        ]

        image_values = [
            ChannelValue(
                x=i, y={"image_value": {"data": "base64Image=="}}, ts=self._TS + i
            )
            for i in range(0, 3)
        ]
        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in numeric_values:
            channels_values_sender.send(
                self._NUMERIC_CHANNEL.name,
                self._NUMERIC_CHANNEL.channelType,
                channel_value,
            )

        for channel_value in text_values:
            channels_values_sender.send(
                self._TEXT_CHANNEL.name, self._TEXT_CHANNEL.channelType, channel_value
            )

        for channel_value in image_values:
            channels_values_sender.send(
                self._IMAGE_CHANNEL.name, self._IMAGE_CHANNEL.channelType, channel_value
            )

        # and
        channels_values_sender.join()

        # then
        # pylint: disable=protected-access
        (args, _) = self._EXPERIMENT._send_channels_values.call_args
        self.assertEqual(len(args), 1)
        self.assertEqual(
            sorted(args[0]),
            sorted(
                [
                    ChannelIdWithValues(
                        channel_id=self._NUMERIC_CHANNEL.id,
                        channel_name=self._NUMERIC_CHANNEL.name,
                        channel_type=self._NUMERIC_CHANNEL.channelType,
                        channel_namespace=ChannelNamespace.USER,
                        channel_values=numeric_values,
                    ),
                    ChannelIdWithValues(
                        channel_id=self._TEXT_CHANNEL.id,
                        channel_name=self._NUMERIC_CHANNEL.name,
                        channel_type=self._NUMERIC_CHANNEL.channelType,
                        channel_namespace=ChannelNamespace.USER,
                        channel_values=text_values,
                    ),
                    ChannelIdWithValues(
                        channel_id=self._IMAGE_CHANNEL.id,
                        channel_name=self._NUMERIC_CHANNEL.name,
                        channel_type=self._NUMERIC_CHANNEL.channelType,
                        channel_namespace=ChannelNamespace.USER,
                        channel_values=image_values,
                    ),
                ]
            ),
        )

    __TIMEOUT = 0.1

    # Using pytest-timeout because Python 2.7 does not support timeouts on acquiring semaphores.
    # Can be refactored once Python 2 support is dropped.
    @pytest.mark.timeout(10 * __TIMEOUT)
    @mock.patch(
        "neptune.internal.channels.channels_values_sender.ChannelsValuesSendingThread._SLEEP_TIME",
        __TIMEOUT,
    )
    def test_send_when_waiting_for_next_value_timed_out(self):
        # given
        numeric_values = [ChannelValue(x=i, y=i, ts=self._TS + i) for i in range(0, 3)]

        # and
        semaphore = threading.Semaphore(0)
        # pylint: disable=protected-access
        self._EXPERIMENT._send_channels_values.side_effect = (
            lambda _: semaphore.release()
        )

        # and
        channels_values_sender = ChannelsValuesSender(experiment=self._EXPERIMENT)

        # when
        for channel_value in numeric_values:
            channels_values_sender.send(
                self._NUMERIC_CHANNEL.name,
                self._NUMERIC_CHANNEL.channelType,
                channel_value,
            )

        # then
        # pylint: disable=protected-access
        semaphore.acquire()
        self._EXPERIMENT._send_channels_values.assert_called_with(
            [
                ChannelIdWithValues(
                    channel_id=self._NUMERIC_CHANNEL.id,
                    channel_name=self._NUMERIC_CHANNEL.name,
                    channel_type=self._NUMERIC_CHANNEL.channelType,
                    channel_namespace=ChannelNamespace.USER,
                    channel_values=numeric_values,
                )
            ]
        )

        # and
        self._EXPERIMENT._send_channels_values.reset_mock()
        channels_values_sender.join()
        # and
        self._EXPERIMENT._send_channels_values.assert_not_called()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .atoms.artifact import Artifact
from .atoms.boolean import Boolean
from .atoms.datetime import Datetime
from .atoms.file import File
from .atoms.float import Float
from .atoms.git_ref import GitRef
from .atoms.integer import Integer
from .atoms.notebook_ref import NotebookRef
from .atoms.run_state import RunState
from .atoms.string import String
from .file_set import FileSet
from .series.file_series import FileSeries
from .series.float_series import FloatSeries
from .series.string_series import StringSeries
from .sets.string_set import StringSet
from .utils import create_attribute_from_type



import os

from setuptools import find_packages, setup

import versioneer

with open(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md"),
    encoding="utf-8",
) as handler:
    long_description = handler.read()


def main():
    root_dir = os.path.dirname(__file__)

    with open(os.path.join(root_dir, "requirements.txt")) as f:
        requirements = [r.strip() for r in f]
        setup(
            name="neptune-client",
            python_requires=">=3.7.0",
            version=versioneer.get_version(),
            description="Neptune Client",
            author="neptune.ai",
            author_email="contact@neptune.ai",
            url="https://neptune.ai/",
            project_urls={
                "Tracker": "https://github.com/neptune-ai/neptune-client/issues",
                "Source": "https://github.com/neptune-ai/neptune-client",
                "Documentation": "https://docs.neptune.ai/",
            },
            long_description=long_description,
            long_description_content_type="text/markdown",
            license="Apache License 2.0",
            install_requires=requirements,
            packages=find_packages(),
            cmdclass=versioneer.get_cmdclass(),
            extras_require={
                "kedro": ["kedro-neptune"],
                "fastai": ["neptune-fastai"],
                "lightgbm": ["neptune-lightgbm"],
                "optuna": ["neptune-optuna"],
                "pytorch-lightning": ["pytorch-lightning"],
                "sacred": ["neptune-sacred"],
                "sklearn": ["neptune-sklearn"],
                "tensorflow-keras": ["neptune-tensorflow-keras"],
                "xgboost": ["neptune-xgboost"],
            },
            entry_points={
                "console_scripts": [
                    "neptune = neptune_cli.main:main",
                ],
            },
            classifiers=[
                # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
                "Development Status :: 5 - Production/Stable",
                "Environment :: Console",
                "Intended Audience :: Developers",
                "Intended Audience :: Science/Research",
                "License :: OSI Approved :: Apache Software License",
                "Natural Language :: English",
                "Operating System :: MacOS",
                "Operating System :: Microsoft :: Windows",
                "Operating System :: POSIX",
                "Operating System :: Unix",
                "Programming Language :: Python :: 3",
                "Programming Language :: Python :: 3.6",
                "Programming Language :: Python :: 3.7",
                "Programming Language :: Python :: 3.8",
                "Programming Language :: Python :: 3.9",
                "Topic :: Software Development :: Libraries :: Python Modules",
                "Programming Language :: Python :: Implementation :: CPython",
                "Topic :: Scientific/Engineering :: Artificial Intelligence",
            ],
            keywords=[
                "MLOps",
                "ML Experiment Tracking",
                "ML Model Registry",
                "ML Model Store",
                "ML Metadata Store",
            ],
        )


if __name__ == "__main__":
    main()



#
# Copyright (c) 2020, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

# pylint: disable=protected-access

import uuid

from neptune.new.exceptions import MetadataInconsistency
from neptune.new.internal.backends.operations_preprocessor import OperationsPreprocessor
from neptune.new.internal.operation import (
    AddStrings,
    AssignFloat,
    AssignString,
    ClearFloatLog,
    ClearImageLog,
    ClearStringSet,
    ConfigFloatSeries,
    DeleteAttribute,
    LogFloats,
    LogImages,
    LogStrings,
    RemoveStrings,
    TrackFilesToArtifact,
    UploadFileSet,
)
from tests.neptune.new.attributes.test_attribute_base import TestAttributeBase

FLog = LogFloats.ValueType
SLog = LogStrings.ValueType
ILog = LogImages.ValueType


class TestOperationsPreprocessor(TestAttributeBase):
    def test_delete_attribute(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process(
            [
                DeleteAttribute(["a"]),
                AssignFloat(["a"], 1),
                DeleteAttribute(["a"]),
                AssignString(["b"], "2"),
                DeleteAttribute(["b"]),
                AssignString(["c"], "2"),
                AssignString(["c"], "3"),
                DeleteAttribute(["c"]),
                DeleteAttribute(["d"]),
                DeleteAttribute(["d"]),
                DeleteAttribute(["e"]),
            ]
        )

        # then
        self.assertEqual(
            processor.get_operations(),
            [
                DeleteAttribute(["a"]),
                AssignString(["b"], "2"),
                DeleteAttribute(["b"]),
                AssignString(["c"], "3"),
                DeleteAttribute(["c"]),
                DeleteAttribute(["d"]),
                DeleteAttribute(["e"]),
            ],
        )
        self.assertEqual(processor.get_errors(), [])

    def test_assign(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process(
            [
                AssignFloat(["a"], 1),
                DeleteAttribute(["a"]),
                AssignString(["a"], "111"),
                DeleteAttribute(["b"]),
                AssignFloat(["b"], 2),
                AssignFloat(["c"], 3),
                AssignString(["c"], "333"),
                AssignString(["d"], "44"),
                AssignFloat(["e"], 5),
                AssignFloat(["e"], 10),
                AssignFloat(["e"], 33),
            ]
        )

        # then
        self.assertEqual(
            processor.get_operations(),
            [
                AssignFloat(["a"], 1),
                DeleteAttribute(["a"]),
                AssignString(["a"], "111"),
                DeleteAttribute(["b"]),
                AssignFloat(["b"], 2),
                AssignFloat(["c"], 3),
                AssignString(["d"], "44"),
                AssignFloat(["e"], 33),
            ],
        )
        self.assertEqual(
            processor.get_errors(),
            [
                MetadataInconsistency(
                    "Cannot perform AssignString operation on c: Attribute is not a String"
                )
            ],
        )

    def test_series(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process(
            [
                LogFloats(["a"], [FLog(1, 2, 3)]),
                ConfigFloatSeries(["a"], min=7, max=70, unit="%"),
                DeleteAttribute(["a"]),
                LogStrings(["a"], [SLog("111", 3, 4)]),
                DeleteAttribute(["b"]),
                LogStrings(["b"], [SLog("222", None, 6)]),
                LogFloats(["c"], [FLog(1, 2, 3)]),
                LogFloats(["c"], [FLog(10, 20, 30), FLog(100, 200, 300)]),
                LogStrings(["d"], [SLog("4", 111, 222)]),
                ClearFloatLog(["e"]),
                LogImages(["f"], [ILog("1", 2, 3)]),
                LogImages(["f"], [ILog("10", 20, 30), FLog("100", 200, 300)]),
                LogImages(["f"], [ILog("1", 2, 3)]),
                LogImages(["f"], [ILog("10", 20, 30), FLog("100", 200, 300)]),
                ClearImageLog(["f"]),
                LogImages(["f"], [ILog("3", 20, 30), FLog("4", 200, 300)]),
                LogImages(["f"], [ILog("5", 2, 3)]),
                LogImages(["f"], [ILog("8", 20, 30), FLog("1000", 200, 300)]),
                LogImages(["g"], [ILog("10", 20, 30), FLog("100", 200, 300)]),
                ClearImageLog(["g"]),
                AssignString(["h"], "44"),
                LogFloats(["h"], [FLog(10, 20, 30), FLog(100, 200, 300)]),
                LogFloats(["i"], [FLog(1, 2, 3)]),
                ConfigFloatSeries(["i"], min=7, max=70, unit="%"),
                ClearFloatLog(["i"]),
                LogFloats(["i"], [FLog(10, 20, 30), FLog(100, 200, 300)]),
            ]
        )

        # then
        self.assertEqual(
            processor.get_operations(),
            [
                LogFloats(["a"], [FLog(1, 2, 3)]),
                DeleteAttribute(["a"]),
                LogStrings(["a"], [FLog("111", 3, 4)]),
                DeleteAttribute(["b"]),
                LogStrings(["b"], [SLog("222", None, 6)]),
                LogFloats(
                    ["c"], [FLog(1, 2, 3), FLog(10, 20, 30), FLog(100, 200, 300)]
                ),
                LogStrings(["d"], [SLog("4", 111, 222)]),
                ClearFloatLog(["e"]),
                ClearImageLog(["f"]),
                LogImages(
                    ["f"],
                    [
                        ILog("3", 20, 30),
                        FLog("4", 200, 300),
                        ILog("5", 2, 3),
                        ILog("8", 20, 30),
                        FLog("1000", 200, 300),
                    ],
                ),
                ClearImageLog(["g"]),
                AssignString(["h"], "44"),
                ClearFloatLog(["i"]),
                LogFloats(["i"], [FLog(10, 20, 30), FLog(100, 200, 300)]),
                ConfigFloatSeries(["i"], min=7, max=70, unit="%"),
            ],
        )
        self.assertEqual(
            processor.get_errors(),
            [
                MetadataInconsistency(
                    "Cannot perform LogFloats operation on h: Attribute is not a Float Series"
                )
            ],
        )

    def test_sets(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process(
            [
                AddStrings(["a"], {"xx", "y", "abc"}),
                DeleteAttribute(["a"]),
                AddStrings(["a"], {"hhh", "gij"}),
                DeleteAttribute(["b"]),
                RemoveStrings(["b"], {"abc", "defgh"}),
                AddStrings(["c"], {"hhh", "gij"}),
                RemoveStrings(["c"], {"abc", "defgh"}),
                AddStrings(["c"], {"qqq"}),
                ClearStringSet(["d"]),
                RemoveStrings(["e"], {"abc", "defgh"}),
                AddStrings(["e"], {"qqq"}),
                ClearStringSet(["e"]),
                AddStrings(["f"], {"hhh", "gij"}),
                RemoveStrings(["f"], {"abc", "defgh"}),
                AddStrings(["f"], {"qqq"}),
                ClearStringSet(["f"]),
                AddStrings(["f"], {"xx", "y", "abc"}),
                RemoveStrings(["f"], {"abc", "defgh"}),
                AssignString(["h"], "44"),
                RemoveStrings(["h"], {""}),
                AssignFloat(["i"], 5),
                AddStrings(["i"], {""}),
            ]
        )

        # then
        self.assertEqual(
            processor.get_operations(),
            [
                AddStrings(["a"], {"xx", "y", "abc"}),
                DeleteAttribute(["a"]),
                AddStrings(["a"], {"hhh", "gij"}),
                DeleteAttribute(["b"]),
                RemoveStrings(["b"], {"abc", "defgh"}),
                AddStrings(["c"], {"hhh", "gij"}),
                RemoveStrings(["c"], {"abc", "defgh"}),
                AddStrings(["c"], {"qqq"}),
                ClearStringSet(["d"]),
                ClearStringSet(["e"]),
                ClearStringSet(["f"]),
                AddStrings(["f"], {"xx", "y", "abc"}),
                RemoveStrings(["f"], {"abc", "defgh"}),
                AssignString(["h"], "44"),
                AssignFloat(["i"], 5),
            ],
        )
        self.assertEqual(
            processor.get_errors(),
            [
                MetadataInconsistency(
                    "Cannot perform RemoveStrings operation on h: Attribute is not a String Set"
                ),
                MetadataInconsistency(
                    "Cannot perform AddStrings operation on i: Attribute is not a String Set"
                ),
            ],
        )

    def test_file_set(self):
        # given
        processor = OperationsPreprocessor()

        # when
        processor.process(
            [
                UploadFileSet(["a"], ["xx", "y", "abc"], reset=False),
                DeleteAttribute(["a"]),
                UploadFileSet(["a"], ["hhh", "gij"], reset=False),
                DeleteAttribute(["b"]),
                UploadFileSet(["b"], ["abc", "defgh"], reset=True),
                UploadFileSet(["c"], ["hhh", "gij"], reset=False),
                UploadFileSet(["c"], ["abc", "defgh"], reset=True),
                UploadFileSet(["c"], ["qqq"], reset=False),
                UploadFileSet(["d"], ["hhh", "gij"], reset=False),
                AssignFloat(["e"], 5),
                UploadFileSet(["e"], [""], reset=False),
            ]
        )

        # then
        self.assertEqual(
            processor.get_operations(),
            [
                UploadFileSet(["a"], ["xx", "y", "abc"], reset=False),
                DeleteAttribute(["a"]),
                UploadFileSet(["a"], ["hhh", "gij"], reset=False),
                DeleteAttribute(["b"]),
                UploadFileSet(["b"], ["abc", "defgh"], reset=True),
                UploadFileSet(["c"], ["abc", "defgh"], reset=True),
                UploadFileSet(["c"], ["qqq"], reset=False),
                UploadFileSet(["d"], ["hhh", "gij"], reset=False),
                AssignFloat(["e"], 5),
            ],
        )
        self.assertEqual(
            processor.get_errors(),
            [
                MetadataInconsistency(
                    "Cannot perform UploadFileSet operation on e: Attribute is not a File Set"
                )
            ],
        )

    def test_artifacts(self):
        # given
        processor = OperationsPreprocessor()
        project_uuid = uuid.uuid4()

        # when
        processor.process(
            [
                TrackFilesToArtifact(["a"], project_uuid, [("dir1/", None)]),
                DeleteAttribute(["a"]),
                TrackFilesToArtifact(["b"], project_uuid, [("dir1/", None)]),
                TrackFilesToArtifact(["b"], project_uuid, [("dir2/dir3/", "dir2/")]),
                TrackFilesToArtifact(["b"], project_uuid, [("dir4/dir5/", "dir4/")]),
                AssignFloat(["c"], 5),
                TrackFilesToArtifact(["c"], project_uuid, [("dir1/", None)]),
                TrackFilesToArtifact(["d"], project_uuid, [("dir2/dir3/", "dir2/")]),
                TrackFilesToArtifact(["d"], project_uuid, [("dir4/", None)]),
                TrackFilesToArtifact(["e"], project_uuid, [("dir1/", None)]),
                TrackFilesToArtifact(["e"], project_uuid, [("dir2/dir3/", "dir2/")]),
                TrackFilesToArtifact(["f"], project_uuid, [("dir1/", None)]),
                TrackFilesToArtifact(["f"], project_uuid, [("dir2/dir3/", "dir2/")]),
                TrackFilesToArtifact(["f"], project_uuid, [("dir4/", None)]),
            ]
        )

        # then
        self.assertEqual(
            processor.get_operations(),
            [
                TrackFilesToArtifact(["a"], project_uuid, [("dir1/", None)]),
                DeleteAttribute(["a"]),
                TrackFilesToArtifact(
                    ["b"],
                    project_uuid,
                    [("dir1/", None), ("dir2/dir3/", "dir2/"), ("dir4/dir5/", "dir4/")],
                ),
                AssignFloat(["c"], 5),
                TrackFilesToArtifact(
                    ["d"], project_uuid, [("dir2/dir3/", "dir2/"), ("dir4/", None)]
                ),
                TrackFilesToArtifact(
                    ["e"], project_uuid, [("dir1/", None), ("dir2/dir3/", "dir2/")]
                ),
                TrackFilesToArtifact(
                    ["f"],
                    project_uuid,
                    [("dir1/", None), ("dir2/dir3/", "dir2/"), ("dir4/", None)],
                ),
            ],
        )
        self.assertEqual(
            processor.get_errors(),
            [
                MetadataInconsistency(
                    "Cannot perform TrackFilesToArtifact operation on c: Attribute is not a Artifact"
                ),
            ],
        )
